/* GTK - The GIMP Toolkit
 * Copyright (C) 1995-1997 Peter Mattis, Spencer Kimball and Josh MacDonald
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the
 * Free Software Foundation, Inc., 59 Temple Place - Suite 330,
 * Boston, MA 02111-1307, USA.
 */

/*
 * Modified by the GTK+ Team and others 1997-2000.  See the AUTHORS
 * file for a list of people on the GTK+ Team.  See the ChangeLog
 * files for a list of changes.  These files are distributed with
 * GTK+ at ftp://ftp.gtk.org/pub/gtk/. 
 */

/*
 * Modified by the Markus Lausser <sgop@users.sourceforge.net>
 */

#include <config.h>
#include <stdlib.h>
#include <string.h>
#include <stdio.h>

#include "gtk/gtklabel.h"
#include "gtk/gtkmain.h"
#include "gtk/gtkmenuitem.h"
#include "gtk/gtkprivate.h"
#include "gtk/gtkwidget.h"
#include "gtk/gtkwindow.h"
#include "gtk/gtkstyle.h"
#include "gtk/gtkvbox.h"
#include "gtk/gtkhseparator.h"
#include "gdk/gdkkeysyms.h"
//#include "gtk/gtkalias.h"

#include "ui_tooltips.h"
#ifdef WIN32
#include "lgtk2corealias.h"
#endif

#define DEFAULT_DELAY 500           /* Default delay in ms */
#define STICKY_DELAY 0              /* Delay before popping up next tip if we're sticky */
#define STICKY_REVERT_DELAY 1000    /* Delay before sticky fancytips revert to normal */

static void gtk_fancytips_class_init        (GtkFancyTipsClass *klass);
static void gtk_fancytips_init              (GtkFancyTips      *fancytips);
static void gtk_fancytips_destroy           (GtkObject        *object);

static void* gtk_fancytips_update_object(GtkFancyTips* tips, gint x, gint y);
static void gtk_fancytips_set_active_object(GtkFancyTips* tips, void* object);
static void gtk_fancytips_unset_tip_window(GtkFancyTips* tips);

static GtkObjectClass* parent_class = NULL;

GType gtk_fancytips_get_type(void)
{
    static GType fancytips_type = 0;
    
    if (!fancytips_type)
    {
        static const GTypeInfo fancytips_info =
            {
                sizeof (GtkFancyTipsClass),
                NULL,		/* base_init */
                NULL,		/* base_finalize */
                (GClassInitFunc) gtk_fancytips_class_init,
                NULL,		/* class_finalize */
                NULL,		/* class_data */
                sizeof (GtkFancyTips),
                0,		/* n_preallocs */
                (GInstanceInitFunc) gtk_fancytips_init,
                0,
            };
    
        fancytips_type = g_type_register_static
            (GTK_TYPE_OBJECT, "GtkFancyTips", &fancytips_info, 0);
    }

    return fancytips_type;
}

static void gtk_fancytips_class_init(GtkFancyTipsClass* class)
{
    GtkObjectClass* object_class = (GtkObjectClass*)class;

    parent_class = g_type_class_peek_parent(class);
    object_class->destroy = gtk_fancytips_destroy;
}

static void gtk_fancytips_init(GtkFancyTips* fancytips)
{
    fancytips->tip_window = NULL;
  
    fancytips->delay = DEFAULT_DELAY;
    fancytips->enabled = TRUE;
    fancytips->timer_tag = 0;
    fancytips->last_popdown.tv_sec = -1;
    fancytips->last_popdown.tv_usec = -1;

    fancytips->tv = NULL;
    fancytips->func = NULL;
    fancytips->object = NULL;
    fancytips->sticky_mode = FALSE;
    fancytips->depth = 0;
    fancytips->pos = GTK_FANCY_POS_MOUSE;
}

static void gtk_fancytips_event_handler(
        GtkWidget* widget, GdkEvent* event,
        GtkFancyTips* tips)
{

    if ((event->type == GDK_LEAVE_NOTIFY || event->type == GDK_ENTER_NOTIFY) &&
        event->crossing.detail == GDK_NOTIFY_INFERIOR)
        return;

    if (event->type != GDK_KEY_PRESS && event->type != GDK_KEY_RELEASE)
    {
        if (gtk_get_event_widget(event) != widget) return;
    }
  
    switch (event->type)
    {
    case GDK_EXPOSE:
        /* do nothing */
        break;
    case GDK_ENTER_NOTIFY:
    {
        GdkEventCrossing* e = (GdkEventCrossing*)event;
        gtk_fancytips_update_object(tips, (gint)e->x, (gint)e->y);
        break;
    }
    case GDK_MOTION_NOTIFY:
    {
        GdkEventMotion* e = (GdkEventMotion*)event;
        gtk_fancytips_update_object(tips, (gint)e->x, (gint)e->y);
        break;
    }
    case GDK_LEAVE_NOTIFY:
    {
        tips->sticky_mode = TRUE;
        gtk_fancytips_set_active_object(tips, NULL);
        break;
    }
    case GDK_KEY_PRESS:
    case GDK_BUTTON_PRESS:
    case GDK_BUTTON_RELEASE:
    case GDK_KEY_RELEASE:
    case GDK_PROXIMITY_IN:
    case GDK_SCROLL:
        tips->sticky_mode = FALSE;
        gtk_fancytips_set_active_object(tips, NULL);
        break;
    default:
        break;
    }
}

/*
  static gboolean
  gtk_fancytips_event_handler_before(GtkWidget* widget, GdkEvent* event, GtkFancyTips* tips) {

  switch (event->type) {
  case GDK_KEY_PRESS: {
  GdkEventKey* e = (GdkEventKey*)event;
  if (e->keyval == GDK_s) {
  printf("'s' pressed\n");
  }
  }
  default:
  break;
  }
  return FALSE;
  }
*/

static void
gtk_fancytips_set_view(GtkFancyTips* tips, GtkTreeView* tv, FancyTipsFunc func)
{
    g_return_if_fail(tips != NULL);
    g_return_if_fail(tv != NULL);
    g_return_if_fail(func != NULL);
    g_return_if_fail(tips->tv == NULL);

    tips->tv = tv;
    tips->func = func;

    /*
      g_signal_connect(G_OBJECT(tv), "event",
      G_CALLBACK(gtk_fancytips_event_handler_before), tips);
    */
    g_signal_connect_after(G_OBJECT(tv), "event-after",
                           G_CALLBACK(gtk_fancytips_event_handler), tips);
    g_signal_connect_swapped(G_OBJECT(tv), "destroy",
                             G_CALLBACK(g_object_unref), tips);
}

GtkFancyTips* gtk_fancytips_new(
        GtkTreeView* tv, FancyTipsFunc func,
        GtkFancyTipsPos pos)
{
    GtkFancyTips* tips = g_object_new(GTK_TYPE_FANCYTIPS, NULL);
    g_object_ref(G_OBJECT(tips));
    gtk_fancytips_set_view(tips, tv, func);
    tips->pos = pos;
    return tips;
}

static void tip_window_display_closed(
        GdkDisplay* display, gboolean was_error,
        GtkFancyTips* fancytips)
{
    (void)display;
    (void)was_error;
    gtk_fancytips_unset_tip_window(fancytips);
}

static void disconnect_tip_window_display_closed(GtkFancyTips* tips)
{
    g_signal_handlers_disconnect_by_func
        (gtk_widget_get_display(tips->tip_window),
         (gpointer)tip_window_display_closed, tips);
}

static void gtk_fancytips_unset_tip_window(GtkFancyTips* tips)
{
    if (tips->tip_window)
    {
        disconnect_tip_window_display_closed(tips);
        gtk_widget_destroy(tips->tip_window);
        tips->tip_window = NULL;
    }
}

static void gtk_fancytips_destroy(GtkObject* object)
{
    GtkFancyTips* fancytips = GTK_FANCYTIPS(object);

    g_return_if_fail(fancytips != NULL);

    if (fancytips->timer_tag)
    {
        g_source_remove(fancytips->timer_tag);
        fancytips->timer_tag = 0;
    }

    gtk_fancytips_unset_tip_window(fancytips);

    GTK_OBJECT_CLASS(parent_class)->destroy(object);
}

static void gtk_fancytips_update_screen(GtkFancyTips* tips, gboolean new_window)
{
    gboolean screen_changed = FALSE;
  
    if (tips->object && tips->tv)
    {
        GdkScreen* screen = gtk_widget_get_screen(GTK_WIDGET(tips->tv));

        screen_changed = (screen != gtk_widget_get_screen(tips->tip_window));

        if (screen_changed)
        {
            if (!new_window) disconnect_tip_window_display_closed(tips);
            gtk_window_set_screen(GTK_WINDOW(tips->tip_window), screen);
        }
    }

    if (screen_changed || new_window)
        g_signal_connect(gtk_widget_get_display(tips->tip_window), "closed",
                         G_CALLBACK(tip_window_display_closed), tips);
}

static gint gtk_fancytips_paint_window(GtkFancyTips* tips)
{
    GtkRequisition req;

    gtk_widget_size_request(tips->tip_window, &req);
    gtk_paint_flat_box(tips->tip_window->style, tips->tip_window->window,
                       GTK_STATE_NORMAL, GTK_SHADOW_OUT, 
                       NULL, GTK_WIDGET(tips->tip_window), "tooltip",
                       0, 0, req.width, req.height);

    return FALSE;
}

void gtk_fancytips_force_window(GtkFancyTips* tips)
{
    g_return_if_fail(GTK_IS_FANCYTIPS(tips));

    if (!tips->tip_window)
    {
        tips->tip_window = gtk_window_new(GTK_WINDOW_POPUP);
        //    gtk_widget_set_size_request(tips->tip_window, 200, 200);
        gtk_fancytips_update_screen(tips, TRUE);
        gtk_widget_set_app_paintable(tips->tip_window, TRUE);
        gtk_window_set_resizable(GTK_WINDOW(tips->tip_window), FALSE);
        gtk_widget_set_name(tips->tip_window, "gtk-tooltips");
        gtk_container_set_border_width(GTK_CONTAINER(tips->tip_window), 5);
    
        g_signal_connect_swapped(tips->tip_window, "expose_event",
                                 G_CALLBACK(gtk_fancytips_paint_window), tips);
    
        g_signal_connect(tips->tip_window, "destroy",
                         G_CALLBACK(gtk_widget_destroyed), &tips->tip_window);
    }
}

void gtk_fancytips_enable(GtkFancyTips* tips)
{
    g_return_if_fail(tips != NULL);
    tips->enabled = TRUE;
}

void gtk_fancytips_disable(GtkFancyTips* tips)
{
    g_return_if_fail(tips != NULL);
    tips->sticky_mode = FALSE;
    gtk_fancytips_set_active_object(tips, NULL);
    tips->enabled = FALSE;
}

void gtk_fancytips_set_delay(GtkFancyTips* tips, guint delay)
{
    g_return_if_fail(tips != NULL);
    tips->delay = delay;
}

static void gtk_fancytips_move_tips(GtkFancyTips* tips)
{
    GtkWidget* widget;
    gint x, y, w, h;
    GdkScreen* screen;
    GdkScreen* pointer_screen;
    gint monitor_num, px, py;
    GdkRectangle monitor;
    GtkRequisition requisition;

    widget = GTK_WIDGET(tips->tv);

    gtk_fancytips_update_screen(tips, FALSE);
    screen = gtk_widget_get_screen(widget);

    gtk_widget_size_request(tips->tip_window, &requisition);
    w = requisition.width;
    h = requisition.height;

    gdk_window_get_origin(widget->window, &x, &y);
    if (GTK_WIDGET_NO_WINDOW(widget))
    {
        x += widget->allocation.x;
        y += widget->allocation.y;
    }

    gdk_display_get_pointer(gdk_screen_get_display(screen),
                            &pointer_screen, &px, &py, NULL);
    if (pointer_screen != screen)
    {
        px = x;
        py = y;
    }
    monitor_num = gdk_screen_get_monitor_at_point(screen, px, py);
    gdk_screen_get_monitor_geometry(screen, monitor_num, &monitor);

    switch (tips->pos)
    {
    case GTK_FANCY_POS_LEFT:
        x -= w;
        y = py - 5;
        break;
    case GTK_FANCY_POS_RIGHT:
        x += widget->allocation.width;
        y = py - 5;
        break;
    case GTK_FANCY_POS_MOUSE:
    default:
        x = px + 10;
        y = py + 10;
        break;
    }

    if (x + w > monitor.x + monitor.width)
    {
        x = monitor.x + monitor.width - w;
        if (y + h > monitor.y + monitor.height)
            y = py - h - 10;
    }
    else if (y + h > monitor.y + monitor.height)
    {
        y = monitor.y + monitor.height - h;
    }

    gtk_window_move(GTK_WINDOW(tips->tip_window), x, y);
}

static void gtk_fancytips_draw_tips(GtkFancyTips* tips)
{
    GtkWidget* widget;
  
    if (!tips->tip_window)
    {
        gtk_fancytips_force_window(tips);
    }
    else if (GTK_WIDGET_VISIBLE(tips->tip_window))
    {
        g_get_current_time(&tips->last_popdown);
    }

    widget = GTK_WIDGET(tips->tv);

    {
        GtkWidget* child;
        child = gtk_bin_get_child(GTK_BIN(tips->tip_window));
        if (child) gtk_widget_destroy(child);
        child = tips->func(tips->object, tips->depth);
        if (child)
        {
            /*
              GtkWidget* vbox = gtk_vbox_new(FALSE, 5);
              GtkWidget* sep = gtk_hseparator_new();
              GtkWidget* label = gtk_label_new("<span size='smaller'>Hit 's' to make this tooltip sticky</span>");
              gtk_widget_show(vbox);
              gtk_widget_show(sep);
              gtk_widget_show(label);
              gtk_label_set_use_markup(GTK_LABEL(label), TRUE);
              gtk_misc_set_alignment(GTK_MISC(label), 1, 0.5);
              gtk_box_pack_start(GTK_BOX(vbox), child, FALSE, FALSE, 0);
              gtk_box_pack_start(GTK_BOX(vbox), sep, FALSE, FALSE, 0);
              gtk_box_pack_start(GTK_BOX(vbox), label, FALSE, FALSE, 0);
              gtk_container_add(GTK_CONTAINER(tips->tip_window), vbox);
            */
            gtk_container_add(GTK_CONTAINER(tips->tip_window), child);
        }
        else
        {
            gtk_widget_hide(tips->tip_window);
            return;
        }
    }

    gtk_fancytips_move_tips(tips);
    gtk_widget_show(tips->tip_window);

}

static gboolean gtk_fancytips_recently_shown(GtkFancyTips* tips)
{
    GTimeVal now;
    glong msec;
  
    g_get_current_time(&now);
    msec = (now.tv_sec  - tips->last_popdown.tv_sec) * 1000 +
        (now.tv_usec - tips->last_popdown.tv_usec) / 1000;
    return (msec < STICKY_REVERT_DELAY);
}

static gint gtk_fancytips_timeout(gpointer data)
{
    GtkFancyTips* tips = (GtkFancyTips*)data;

    if (tips->timer_tag)
    {
        g_source_remove(tips->timer_tag);
        tips->timer_tag = 0;
    }

    if (tips->object && GTK_WIDGET_DRAWABLE(tips->tv))
        gtk_fancytips_draw_tips(tips);

    return FALSE;
}

static void gtk_fancytips_set_active_object(GtkFancyTips* tips, void* object)
{
    void* old_object = tips->object;

    if (!tips->enabled) return;
  
    tips->object = object;

    if (!old_object || tips->timer_tag)
    {
        // tips not shown or about to show.
        if (tips->timer_tag)
        {
            g_source_remove(tips->timer_tag);
            tips->timer_tag = 0;
        }
        if (tips->object)
        {
            if (tips->sticky_mode && gtk_fancytips_recently_shown(tips))
                tips->timer_tag = g_timeout_add(STICKY_DELAY, gtk_fancytips_timeout, tips);
            else
                tips->timer_tag = g_timeout_add(tips->delay, gtk_fancytips_timeout, tips);
        }
    }
    else if (tips->object)
    {
        if (old_object == tips->object)
            gtk_fancytips_move_tips(tips);
        else
            gtk_fancytips_draw_tips(tips);
    }
    else
    {
        if (GTK_WIDGET_VISIBLE(tips->tip_window))
            g_get_current_time(&tips->last_popdown);
        gtk_widget_hide(tips->tip_window);
    }
}

static void* gtk_fancytips_update_object(GtkFancyTips* tips, gint x, gint y)
{
    void* object = NULL;
    GtkTreePath* path = NULL;

    if (gtk_tree_view_get_path_at_pos(tips->tv, x, y, &path, NULL, NULL, NULL))
    {
        GtkTreeModel* model = gtk_tree_view_get_model(tips->tv);
        GtkTreeIter iter;
        if (gtk_tree_model_get_iter(model, &iter, path))
        {
            tips->depth = gtk_tree_path_get_depth(path);
            gtk_tree_model_get(model, &iter, 0, &object, -1);
        }
        gtk_tree_path_free(path);
    }

    if (!object) tips->sticky_mode = TRUE;
    gtk_fancytips_set_active_object(tips, object);
    return NULL; //FIXME
}

#ifdef WIN32
#define _UI_TOOLTIPS_C_
#include "lgtk2corealiasdef.c"
#endif
