/* Copyright (C) 2005-2007 sgop@users.sourceforge.net
 * This is free software distributed under the terms of the
 * GNU Public License.  See the file COPYING for details.
 */
/* $Revision: 1.14 $
 * $Date: 2007/04/29 09:27:29 $
 * $Author: sgop $
 */

#include <string.h>

#include "ui.h"
#include "module.h"
#include "xml_handler.h"
#include "plugin.h"
#include "dynamic.h"
#include "legacy.h"
#ifdef WIN32
#include "lcorealias.h"
#endif

static ui_interface_t UI;

typedef struct {
    module_t _module;
    void (*init)(ui_interface_t*);
    void (*run)(void);
    void (*done)(void);
} ui_module_t;

static char* TheUI = NULL;
static gboolean UISet = FALSE;
static ui_module_t* UIModule = NULL;
static GMainLoop* MainLoop = NULL;


static void ui_run_daemon(void)
{
    if (!MainLoop) MainLoop = g_main_loop_new(NULL, FALSE);
    g_main_loop_run(MainLoop);
}

static void ui_quit_daemon(void)
{
    if (MainLoop)
    {
        g_main_loop_quit(MainLoop);
        g_main_loop_unref(MainLoop);
    }
}

#define CHECK_SYMBOL(sym)                       \
    if (!(sym))                                                         \
    {                                                                   \
        g_log("UI", G_LOG_LEVEL_WARNING, "could not find symbol '%s'", #sym); \
        lopster_dll_close(handle);                                      \
        g_assert(FALSE);                                                \
        return FALSE;                                                   \
    }

static gboolean ui_load_syms(void)
{
    const char* filename;
    GModule* handle;
    
    UI.enter_main_loop = ui_run_daemon;
    UI.leave_main_loop = ui_quit_daemon;
    
    if (!g_module_supported())
    {
        g_log("UI", G_LOG_LEVEL_WARNING, "plugins not supported by system");
        return FALSE;
    }
    
    filename = plugin_find_ui_library("core");
    if (!filename)
    {
        g_log("UI", G_LOG_LEVEL_DEBUG, "running in deamon mode");
        ui_set_name(NULL);
        return FALSE;
    }
    
    // RTLD_GLOBAL: allow plugin to export symbols, since the
    // user interface of other plugins might want to access
    handle = lopster_dll_open(filename, G_MODULE_BIND_LAZY);
    if (!handle)
    {
        g_log("UI", G_LOG_LEVEL_WARNING, "%s", g_module_error());
        g_log("UI", G_LOG_LEVEL_DEBUG, "running in deamon mode");
        return FALSE;
    }
    
    CHECK_SYMBOL(g_module_symbol(handle, "ui_init", (void**)&UIModule->init));
    CHECK_SYMBOL(g_module_symbol(handle, "ui_run", (void**)&UIModule->run));
    CHECK_SYMBOL(g_module_symbol(handle, "ui_done", (void**)&UIModule->done));
    
    return TRUE;
}



static void xml_load_ui(xmlDocPtr xfile, xmlNodePtr xroot)
{
    xmlChar* attr = xmlGetProp(xroot, "Name");
    (void)xfile;
    ui_set_name(attr);
    if (attr) xmlFree(attr);
}  

static void xml_save_ui(xmlDocPtr xfile, xmlNodePtr xroot)
{
    (void)xfile;
    xmlNewProp(xroot, "Name", TheUI?TheUI:"");
}

/**
 * Get the name of the ui 
 * @return the name of the ui
 */
const char* ui_get_name()
{
    return TheUI;
}

void ui_set_name(const char* ui)
{
    g_free(TheUI);
    TheUI = g_strdup(ui);
    UISet = TRUE;
    // only save if user changes it.
    if (UIModule && module_running(MODULE(UIModule)))
        lopster_xml_save("core:ui", 1000);
}



void UI_enter_main_loop(void) {
    if (UI.enter_main_loop) UI.enter_main_loop();
}
void UI_leave_main_loop(void) {
    if (UI.leave_main_loop) UI.leave_main_loop();
}
void UI_download_show_details(download_t* download) {
    if (UI.download_show_details) UI.download_show_details(download);
}
void UI_file_tree_update_label(file_tree_t* tree , const char* str) {
    if (UI.file_tree_update_label) UI.file_tree_update_label(tree, str);
}
void UI_file_tree_set_progress(file_tree_t* tree, double prog) {
    if (UI.file_tree_set_progress) UI.file_tree_set_progress(tree, prog);
}
void UI_file_tree_set_progress_name(file_tree_t* tree, const char* str) {
    if (UI.file_tree_set_progress_name) UI.file_tree_set_progress_name(tree, str);
}
void UI_file_tree_set_abort(file_tree_t* tree, GFunc func, void* data) {
    if (UI.file_tree_set_abort) UI.file_tree_set_abort(tree, func, data);
}
void UI_lib_working(gboolean val) {
    if (UI.lib_working) UI.lib_working(val);
}
void UI_chat_switch_page(chat_page_t* page) {
    if (UI.chat_switch_page) UI.chat_switch_page(page);
}
gboolean UI_chat_page_printable(chat_page_t* page) {
    if (UI.chat_page_printable) return UI.chat_page_printable(page);
    else return TRUE;
}
void UI_chat_create_simple_page(chat_page_t* page) {
    if (UI.chat_create_simple_page) UI.chat_create_simple_page(page);
}
void UI_chat_print(chat_page_t* page, int index, va_list ap) {
    if (UI.chat_print) UI.chat_print(page, index, ap);
}
void UI_threads_show(const char* message) {
    if (UI.threads_show) UI.threads_show(message);
}
#ifdef OWN_SOCKET_TRHEAD
void UI_thread_enter() {
    if (UI.thread_enter) UI.thread_enter();
}
void UI_thread_leave() {
    if (UI.thread_leave) UI.thread_leave();
}
#endif

static void ui_init() {
    if (UIModule->init) UIModule->init(&UI);
}
static void ui_run() {
    if (UIModule->run) UIModule->run();
}
static void ui_done() {
    if (UIModule->done) UIModule->done();
}

static void ui_module_init(ui_module_t* module)
{
    module_init(MODULE(module), "User Interface Module",
                (ModuleFunc)ui_init, (ModuleFunc)ui_run, (ModuleFunc)ui_done);
    module->init = NULL;
    module->run = NULL;
    module->done = NULL;
    
    module_require(xml_module_instance);
    lopster_xml_register_home
        ("core:ui", "ui.xml", "UserInterface", xml_load_ui, xml_save_ui);
    // if both no command line argument and no saved setting is available
    // set the gtk2 ui to make new users happy.
    if (!UISet && !lopster_xml_load("core:ui")) ui_set_name("gtk2");
    module_require(dynamic_module_instance);
    
    ui_load_syms();
    
    // require some other modules
    //FIXME: should we really require them all here??
    module_require(download_module_instance);
    module_require(upload_module_instance);
    module_require(library_module_instance);
    module_require(legacy_module_instance);
}

module_t* ui_module_instance()
{
    if (!UIModule)
    {
        UIModule = g_malloc0(sizeof(*UIModule));
        ui_module_init(UIModule);
    }
    return MODULE(UIModule);
}

#ifdef WIN32
#define _UI_C_
#include "lcorealiasdef.c"
#endif
